@@ -1,7 +1,9 @@
 #This file is part of Tryton.  The COPYRIGHT file at the top level of
 #this repository contains the full copyright notices and license terms.
 from trytond.model import ModelSQL, ModelView, fields
-from trytond.tools import Cache, reduce_ids
+from trytond.tools import reduce_ids
+from trytond.backend import TableHandler
+from trytond.pyson import Not, Equal, Eval, If, Bool, In
 import uuid
 import vobject
 import dateutil.tz
@@ -20,7 +22,7 @@ class Todo(ModelSQL, ModelView):
     _rec_name = 'uuid'
 
     calendar = fields.Many2One('calendar.calendar', 'Calendar',
-            required=True, select=1)
+            required=True, select=1, ondelete="CASCADE")
     alarms = fields.One2Many('calendar.todo.alarm', 'todo', 'Alarms')
     classification = fields.Selection([
         ('public', 'Public'),
@@ -29,34 +31,42 @@ class Todo(ModelSQL, ModelView):
         ], 'Classification', required=True)
     completed = fields.DateTime('Completed',
             states={
-                'readonly': "(status != 'completed')",
+                'readonly': Not(Equal(Eval('status'), 'completed')),
             }, depends=['status'])
     description = fields.Text('Description')
     dtstart = fields.DateTime('Start Date', select=1)
     location = fields.Many2One('calendar.location', 'Location')
     organizer = fields.Char('Organizer', states={
-        'required': "bool(attendees) and not bool(parent)",
+        'required': If(Bool(Eval('attendees')),
+            Not(Bool(Eval('parent'))),
+            False),
         }, depends=['attendees', 'parent'])
     attendees = fields.One2Many('calendar.todo.attendee', 'todo',
             'Attendees')
     percent_complete = fields.Integer('Percent complete',
             states={
-                'readonly': "(status not in ('needs-action', 'in-process'))",
+                'readonly': Not(In(Eval('status'),
+                    ['needs-action', 'in-process'])),
             }, depends=['status'])
-    recurrences = fields.One2Many('calendar.todo', 'parent', 'Recurrences',
-            domain=["('uuid', '=', uuid)",
-                "('calendar', '=', calendar)"],
+    occurences = fields.One2Many('calendar.todo', 'parent', 'Occurences',
+            domain=[
+                ('uuid', '=', Eval('uuid')),
+                ('calendar', '=', Eval('calendar')),
+            ],
             states={
-                'invisible': "bool(parent)",
+                'invisible': Bool(Eval('parent')),
             }, depends=['uuid', 'calendar', 'parent'])
     recurrence = fields.DateTime('Recurrence', select=1, states={
-                'invisible': "not bool(parent)",
+                'invisible': Not(Bool(Eval('_parent_parent'))),
+                'required': Bool(Eval('_parent_parent')),
                 }, depends=['parent'])
     sequence = fields.Integer('Sequence')
     parent = fields.Many2One('calendar.todo', 'Parent',
-            domain=["('uuid', '=', uuid)",
-                "('parent', '=', False)",
-                "('calendar', '=', calendar)"],
+            domain=[
+                ('uuid', '=', Eval('uuid')),
+                ('parent', '=', False),
+                ('calendar', '=', Eval('calendar'))
+            ],
             ondelete='CASCADE', depends=['uuid', 'calendar'])
     timezone = fields.Selection('timezones', 'Timezone')
     status = fields.Selection([
@@ -74,32 +84,26 @@ class Todo(ModelSQL, ModelView):
             'todo', 'category', 'Categories')
     exdates = fields.One2Many('calendar.todo.exdate', 'todo', 'Exception Dates',
             states={
-                'invisible': "bool(parent)",
+                'invisible': Bool(Eval('parent')),
             }, depends=['parent'])
     exrules = fields.One2Many('calendar.todo.exrule', 'todo', 'Exception Rules',
             states={
-                'invisible': "bool(parent)",
+                'invisible': Bool(Eval('parent')),
             }, depends=['parent'])
     rdates = fields.One2Many('calendar.todo.rdate', 'todo', 'Recurrence Dates',
             states={
-                'invisible': "bool(parent)",
+                'invisible': Bool(Eval('parent')),
             }, depends=['parent'])
     rrules = fields.One2Many('calendar.todo.rrule', 'todo', 'Recurrence Rules',
             states={
-                'invisible': "bool(parent)",
+                'invisible': Bool(Eval('parent')),
             }, depends=['parent'])
-    calendar_owner = fields.Function('get_calendar_field',
-            type='many2one', relation='res.user', string='Owner',
-            fnct_search='search_calendar_field')
-    calendar_read_users = fields.Function('get_calendar_field',
-            type='many2many', relation='res.user', string='Read Users',
-            fnct_search='search_calendar_field')
-    calendar_write_users = fields.Function('get_calendar_field',
-            type='many2many', relation='res.user', string='Write Users',
-            fnct_search='search_calendar_field')
-    classification_public = fields.Function('get_classification_public',
-            type='boolean', string='Classification Public',
-            fnct_search='search_classification_public')
+    calendar_owner = fields.Function(fields.Many2One('res.user', 'Owner'),
+            'get_calendar_field', searcher='search_calendar_field')
+    calendar_read_users = fields.Function(fields.One2Many('res.user', None,
+        'Read Users'), 'get_calendar_field', searcher='search_calendar_field')
+    calendar_write_users = fields.Function(fields.One2Many('res.user', None,
+        'Write Users'), 'get_calendar_field', searcher='search_calendar_field')
     vtodo = fields.Binary('vtodo')
 
     def __init__(self):
@@ -115,6 +119,20 @@ class Todo(ModelSQL, ModelView):
             'invalid_recurrence': 'Recurrence can not be recurrent!',
         })
 
+    def init(self, cursor, module_name):
+        # Migrate from 1.4: remove classification_public
+        model_data_obj = self.pool.get('ir.model.data')
+        rule_obj = self.pool.get('ir.rule')
+        model_data_ids = model_data_obj.search(cursor, 0, [
+            ('fs_id', '=', 'rule_group_read_todo_line3'),
+            ('module', '=', module_name),
+            ('inherit', '=', False),
+            ], limit=1)
+        if model_data_ids:
+            model_data = model_data_obj.browse(cursor, 0, model_data_ids[0])
+            rule_obj.delete(cursor, 0, model_data.db_id)
+        return super(Todo, self).init(cursor, module_name)
+
     def default_uuid(self, cursor, user, context=None):
         return str(uuid.uuid4())
 
@@ -129,7 +147,7 @@ class Todo(ModelSQL, ModelView):
         user_ = user_obj.browse(cursor, user, user, context=context)
         return user_.timezone
 
-    def on_change_status(self, cursor, user, ids, vals, context=None):
+    def on_change_status(self, cursor, user, vals, context=None):
         res = {}
         if 'status' not in vals:
             return res
@@ -143,7 +161,7 @@ class Todo(ModelSQL, ModelView):
     def timezones(self, cursor, user, context=None):
         return [(x, x) for x in pytz.common_timezones] + [('', '')]
 
-    def get_calendar_field(self, cursor, user, ids, name, arg, context=None):
+    def get_calendar_field(self, cursor, user, ids, name, context=None):
         assert name in ('calendar_owner', 'calendar_read_users',
                 'calendar_write_users'), 'Invalid name'
         res = {}
@@ -155,35 +173,8 @@ class Todo(ModelSQL, ModelView):
                 res[todo.id] = todo.calendar[name].id
         return res
 
-    def search_calendar_field(self, cursor, user, name, args, context=None):
-        args2 = []
-        i = 0
-        while i < len(args):
-            field = args[i][0][9:]
-            args2.append(tuple(['calendar.' + field] + list(args[i])[1:]))
-            i += 1
-        return args2
-
-    def get_classification_public(self, cursor, user, ids, name, arg,
-            context=None):
-        res = {}
-        for todo in self.browse(cursor, user, ids, context=context):
-            res[todo.id] = False
-            if todo.classification == 'public':
-                res[todo.id] = True
-        return res
-
-    def search_classification_public(self, cursor, user, name, args,
-            context=None):
-        args2 = []
-        i = 0
-        while i < len(args):
-            if args[i][2]:
-                args2.append(('classification', '=', 'public'))
-            else:
-                args2.append(('classification', '!=', 'public'))
-            i += 1
-        return args2
+    def search_calendar_field(self, cursor, user, name, clause, context=None):
+        return [('calendar.' + name[9:],) + clause[1:]]
 
     def check_recurrence(self, cursor, user, ids):
         '''
@@ -196,7 +187,7 @@ class Todo(ModelSQL, ModelView):
                     or todo.rrules \
                     or todo.exdates \
                     or todo.exrules \
-                    or todo.recurrences:
+                    or todo.occurences:
                 return False
         return True
 
@@ -226,10 +217,10 @@ class Todo(ModelSQL, ModelView):
                     for calendar_id in calendar_ids:
                         new_id = self.copy(cursor, 0, todo.id, default={
                             'calendar': calendar_id,
-                            'recurrences': False,
+                            'occurences': False,
                             }, context=context)
-                        for recurrence in todo.recurrences:
-                            self.copy(cursor, 0, recurrence.id, default={
+                        for occurence in todo.occurences:
+                            self.copy(cursor, 0, occurence.id, default={
                                 'calendar': calendar_id,
                                 'parent': new_id,
                                 }, context=context)
@@ -338,10 +329,10 @@ class Todo(ModelSQL, ModelView):
                         for calendar_id in calendar_ids:
                             new_id = self.copy(cursor, 0, todo.id, default={
                                 'calendar': calendar_id,
-                                'recurrences': False,
+                                'occurences': False,
                                 }, context=context)
-                            for recurrence in todo.recurrences:
-                                self.copy(cursor, 0, recurrence.id, default={
+                            for occurence in todo.occurences:
+                                self.copy(cursor, 0, occurence.id, default={
                                     'calendar': calendar_id,
                                     'parent': new_id,
                                     }, context=context)
@@ -643,31 +634,31 @@ class Todo(ModelSQL, ModelView):
 
         res['vtodo'] = vtodo.serialize()
 
-        recurrences_todel = []
+        occurences_todel = []
         if todo:
-            recurrences_todel = [x.id for x in todo.recurrences]
+            occurences_todel = [x.id for x in todo.occurences]
         for vtodo in vtodos:
             todo_id = None
             if todo:
-                for recurrence in todo.recurrences:
-                    if recurrence.recurrence.replace(tzinfo=tzlocal) \
+                for occurence in todo.occurences:
+                    if occurence.recurrence.replace(tzinfo=tzlocal) \
                             == vtodo.recurrence_id.value:
-                        todo_id = recurrence.id
-                        recurrences_todel.remove(recurrence.id)
+                        todo_id = occurence.id
+                        occurences_todel.remove(occurence.id)
             vals = self.ical2values(cursor, user, todo_id, ical,
                     calendar_id, vtodo=vtodo, context=context)
             if todo:
                 vals['uuid'] = todo.uuid
             else:
                 vals['uuid'] = res['uuid']
-            res.setdefault('recurrences', [])
+            res.setdefault('occurences', [])
             if todo_id:
-                res['recurrences'].append(('write', todo_id, vals))
+                res['occurences'].append(('write', todo_id, vals))
             else:
-                res['recurrences'].append(('create', vals))
-        if recurrences_todel:
-            res.setdefault('recurrences', [])
-            res['recurrences'].append(('delete', recurrences_todel))
+                res['occurences'].append(('create', vals))
+        if occurences_todel:
+            res.setdefault('occurences', [])
+            res['occurences'].append(('delete', occurences_todel))
         return res
 
     def todo2ical(self, cursor, user, todo, context=None):
@@ -762,14 +753,13 @@ class Todo(ModelSQL, ModelView):
         if not hasattr(vtodo, 'last-modified'):
             vtodo.add('last-modified')
         vtodo.last_modified.value = date.replace(tzinfo=tzlocal).astimezone(tztodo)
-        if todo.recurrence:
+        if todo.recurrence and todo.parent:
             if not hasattr(vtodo, 'recurrence-id'):
                 vtodo.add('recurrence-id')
-            if todo.all_day:
-                vtodo.recurrence_id.value = todo.recurrence.date()
-            else:
-                vtodo.recurrence_id.value = todo.recurrence\
-                        .replace(tzinfo=tzlocal).astimezone(tztodo)
+            vtodo.recurrence_id.value = todo.recurrence\
+                    .replace(tzinfo=tzlocal).astimezone(tztodo)
+        elif hasattr(vtodo, 'recurrence-id'):
+            del vtodo.recurrence_id
         if todo.status:
             if not hasattr(vtodo, 'status'):
                 vtodo.add('status')
@@ -844,8 +834,8 @@ class Todo(ModelSQL, ModelView):
             if valarm:
                 vtodo.valarm_list.append(valarm)
 
-        for recurrence in todo.recurrences:
-            rical = self.todo2ical(cursor, user, recurrence, context=context)
+        for occurence in todo.occurences:
+            rical = self.todo2ical(cursor, user, occurence, context=context)
             ical.vtodo_list.append(rical.vtodo)
         return ical
 
@@ -866,17 +856,26 @@ TodoCategory()
 
 
 class TodoRDate(ModelSQL, ModelView):
-    'Recurrence Date'
+    'Todo Recurrence Date'
     _description = __doc__
     _name = 'calendar.todo.rdate'
-    _inherits = {'calendar.rdate': 'calendar_rdate'}
+    _inherits = {'calendar.date': 'calendar_date'}
     _rec_name = 'datetime'
 
-    calendar_rdate = fields.Many2One('calendar.rdate', 'Calendar RDate',
+    calendar_date = fields.Many2One('calendar.date', 'Calendar Date',
             required=True, ondelete='CASCADE', select=1)
     todo = fields.Many2One('calendar.todo', 'Todo', ondelete='CASCADE',
             select=1, required=True)
 
+    def init(self, cursor, module_name):
+        # Migration from 1.4: calendar_rdate renamed to calendar_date
+        table = TableHandler(cursor, self, module_name)
+        old_column = 'calendar_rdate'
+        if table.column_exist(old_column):
+            table.column_rename(old_column, 'calendar_date')
+
+        return super(TodoRDate, self).init(cursor, module_name)
+
     def create(self, cursor, user, values, context=None):
         todo_obj = self.pool.get('calendar.todo')
         if values.get('todo'):
@@ -901,11 +900,11 @@ class TodoRDate(ModelSQL, ModelView):
 
     def delete(self, cursor, user, ids, context=None):
         todo_obj = self.pool.get('calendar.todo')
-        rdate_obj = self.pool.get('calendar.rdate')
+        rdate_obj = self.pool.get('calendar.date')
         if isinstance(ids, (int, long)):
             ids = [ids]
         todo_rdates = self.browse(cursor, user, ids, context=context)
-        rdate_ids = [a.calendar_rdate.id for a in todo_rdates]
+        rdate_ids = [a.calendar_date.id for a in todo_rdates]
         todo_ids = [x.todo.id for x in todo_rdates]
         if todo_ids:
             # Update write_date of todo
@@ -916,15 +915,15 @@ class TodoRDate(ModelSQL, ModelView):
         return res
 
     def _date2update(self, cursor, user, date, context=None):
-        date_obj = self.pool.get('calendar.rdate')
+        date_obj = self.pool.get('calendar.date')
         return date_obj._date2update(cursor, user, date, context=context)
 
     def date2values(self, cursor, user, date, context=None):
-        date_obj = self.pool.get('calendar.rdate')
+        date_obj = self.pool.get('calendar.date')
         return date_obj.date2values(cursor, user, date, context=context)
 
     def date2date(self, cursor, user, date, context=None):
-        date_obj = self.pool.get('calendar.rdate')
+        date_obj = self.pool.get('calendar.date')
         return date_obj.date2date(cursor, user, date, context=context)
 
 TodoRDate()
